@@ -66,40 +66,64 @@
 #include "http_config.h"
 #include "http_log.h"
 
-#if SSLEAY_VERSION_NUMBER < 0x600
-#define ERR_print_errors_fp	ERR_print_errors
+/*
+Useful bits for dbx (useful for Ben, at least):
+
+r -d /u/ben/apache/apache_1.2.5-ssl/SSLconf -X
+use /u/ben/work/scuzzy-ssleay8/ssl /u/ben/work/scuzzy-ssleay8/crypto/bio
+*/
+
+#define SSL_IS_OPTIONAL		FALSE	/* Set to TRUE if you want to allow SSL to be disabled on a per-request basis (useful for subrequests) */
+
+#if SSL_IS_OPTIONAL
+# define OPTIONAL_SSL	(!r->connection->client->ssl)
+#else
+# define OPTIONAL_SSL	FALSE
 #endif
 
 #if SSLEAY_VERSION_NUMBER < 0x0800
-#define X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY \
+# error "SSLeay versions below 0.8.x are no longer supported"
+/* But note, they nearly are, if anyone cares */
+#endif
+
+#if SSLEAY_VERSION_NUMBER >= 0x0800
+/* Do session caching - only available with SSLeay 0.8.x */
+# define CACHE_SESSIONS		TRUE
+#endif
+
+/* define this true when SSLeay can do an SSL_free without falling over */
+#define FREE_SESSION	FALSE
+
+/* define this to debug the verification stuff */
+#define DEBUG_VERIFY	FALSE
+
+#if CACHE_SESSIONS
+# include "gcache.h"
+static const char *szCacheServerPort;
+static char *szCacheServerPath;
+
+#define DEBUG_SESSIONS		0
+
+#endif
+
+#if SSLEAY_VERSION_NUMBER < 0x900
+/* correct for a typo */
+# define SSL3_TXT_ADH_DES_192_CBC_SHA SSL3_TXT_ADH_DES_196_CBC_SHA
+#endif
+
+#if SSLEAY_VERSION_NUMBER < 0x600
+# define ERR_print_errors_fp	ERR_print_errors
+#endif
+
+#if SSLEAY_VERSION_NUMBER < 0x0800
+# define X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY \
 		VERIFY_ERR_UNABLE_TO_GET_ISSUER
-#define X509_V_OK VERIFY_OK
-#define SSL_CTX_set_default_verify_paths(ctx) SSL_set_default_verify_paths(ctx)
-#define SSL_CTX_load_verify_locations(ctx,cafile,capath) \
+# define X509_V_OK VERIFY_OK
+# define SSL_CTX_set_default_verify_paths(ctx)	\
+		SSL_set_default_verify_paths(ctx)
+# define SSL_CTX_load_verify_locations(ctx,cafile,capath) \
 		 SSL_load_verify_locations(ctx,cafile,capath)
-#define X509_verify_cert_error_string(error) X509_cert_verify_error_string(error)
-#endif
-
-#if !defined(SSL_TXT_NULL)
-/* text strings for the ciphers */
-#define SSL_TXT_NULL_WITH_MD5		SSL2_TXT_NULL_WITH_MD5
-#define SSL_TXT_RC4_128_WITH_MD5	SSL2_TXT_RC4_128_WITH_MD5
-#define SSL_TXT_RC4_128_EXPORT40_WITH_MD5	SSL2_TXT_RC4_128_EXPORT40_WITH_MD5
-#define SSL_TXT_RC2_128_CBC_WITH_MD5	SSL2_TXT_RC2_128_CBC_WITH_MD5
-#define SSL_TXT_RC2_128_CBC_EXPORT40_WITH_MD5	SSL2_TXT_RC2_128_CBC_EXPORT40_WITH_MD5
-#define SSL_TXT_IDEA_128_CBC_WITH_MD5	SSL2_TXT_IDEA_128_CBC_WITH_MD5
-#define SSL_TXT_DES_64_CBC_WITH_MD5	SSL2_TXT_DES_64_CBC_WITH_MD5
-#define SSL_TXT_DES_64_CBC_WITH_SHA	SSL2_TXT_DES_64_CBC_WITH_SHA
-#define SSL_TXT_DES_192_EDE3_CBC_WITH_MD5	SSL2_TXT_DES_192_EDE3_CBC_WITH_MD5
-#define SSL_TXT_DES_192_EDE3_CBC_WITH_SHA	SSL2_TXT_DES_192_EDE3_CBC_WITH_SHA
-
-#define SSL_TXT_DES_64_CFB64_WITH_MD5_1	SSL2_TXT_DES_64_CFB64_WITH_MD5_1
-#define SSL_TXT_NULL			SSL2_TXT_NULL
-#endif
-
-#ifndef SSL_TXT_DES_64_CFB64_WITH_MD5_1
-/* SSLeay 0.8.0 misses this definition */
-#define SSL_TXT_DES_64_CFB64_WITH_MD5_1	SSL2_TXT_DES_64_CFB64_WITH_MD5_1
+# define X509_verify_cert_error_string(error) X509_cert_verify_error_string(error)
 #endif
 
 typedef enum
@@ -130,6 +154,8 @@ typedef struct
     SSL_CTX *pSSLCtx;
 
     BOOL bFakeBasicAuth;
+
+    time_t tSessionCacheTimeout;
     } SSLConfigRec;
 
 typedef struct
@@ -144,7 +170,7 @@ typedef struct
 
 extern module ssl_module;
 
-static conn_rec *SSLVerifyConn;
+static conn_rec *g_pCurrentConnection;
 
 static const char *SSLRequireCipher(cmd_parms *cmd,SSLDirConfigRec *rec,char *cipher)
     {
@@ -167,7 +193,7 @@ static int SSLCheckCipher(request_rec *r)
       get_module_config(r->server->module_config,&ssl_module);
 
     /* Check to see if SSL is on */
-    if(pConfig->bDisabled)
+    if(pConfig->bDisabled || OPTIONAL_SSL)
 	return DECLINED;
 
     cipher=SSL_get_cipher(r->connection->client->ssl);
@@ -180,6 +206,7 @@ static int SSLCheckCipher(request_rec *r)
 	log_reason(buf,r->filename,r);
 	return FORBIDDEN;
 	}
+
     if(table_get(rec->tbRequiredCiphers,cipher))
 	return OK;
 
@@ -196,49 +223,119 @@ static int SSLCheckCipher(request_rec *r)
 	}
     }
 
-int SSLFixups (request_rec *r)
+static void StrUpper(char *s)
+    {
+    for( ; *s ; ++s)
+	*s=toupper(*s);
+    }
+
+void ExpandCert(table *pEnv,char *szPrefix,char *szCert)
+    {
+    /* Use ap_pstrintf when we get to 1.3 */
+    char var[HUGE_STRING_LEN];
+    char buf[HUGE_STRING_LEN];
+    char *s;
+    /* Expand a X509_oneline entry into it's base components and register
+       them as environment variables. Needed if you want to pass certificate
+       information to CGI's. The naming convention SHOULD be fairly compatible
+       with CGI's written for stronghold's certificate info  - Q */
+    /* FIXME - strtok() and strcspn() may cause problems on some systems - Q */
+
+    sprintf(var,"%sDN",szPrefix);
+    table_set(pEnv,var,szCert);
+    
+    strcpy(buf,szCert);
+    for(s=strtok(buf,"/") ; s != NULL ; s=strtok(NULL,"/"))
+	{
+	int n=strcspn(s,"=");
+	s[n]='\0';
+	StrUpper(s);
+	sprintf(var,"%s%s",szPrefix,s);
+	table_set(pEnv,var,s+n+1);
+	}
+    }
+
+static struct
+    {
+    char *szName;
+    int nKeySize;
+    int nSecretKeySize;
+    } aCipherspecs[]=
+    {
+#if SSLEAY_VERSION_NUMBER >= 0x0900
+    { SSL3_TXT_RSA_IDEA_128_SHA		/*"IDEA-CBC-SHA"*/,	128, 128 },
+#endif
+    { SSL3_TXT_RSA_NULL_MD5		/*"NULL-MD5"*/, 	0, 0 },
+    { SSL3_TXT_RSA_NULL_SHA	 	/*"NULL-SHA"*/, 	0, 0 },
+    { SSL3_TXT_RSA_RC4_40_MD5		/*"EXP-RC4-MD5"*/,	128, 40 },
+    { SSL3_TXT_RSA_RC4_128_MD5		/*"RC4-MD5"*/,		128, 128 },
+    { SSL3_TXT_RSA_RC4_128_SHA		/*"RC4-SHA"*/,		128, 128 },
+    { SSL3_TXT_RSA_RC2_40_MD5		/*"EXP-RC2-CBC-MD5"*/,	128, 40 },
+    { SSL3_TXT_RSA_IDEA_128_SHA		/*"IDEA-CBC-MD5"*/,	128, 128 },
+    { SSL3_TXT_RSA_DES_40_CBC_SHA	/*"EXP-DES-CBC-SHA"*/,	56, 40 },
+    { SSL3_TXT_RSA_DES_64_CBC_SHA	/*"DES-CBC-SHA"*/,	56, 56 },
+    { SSL3_TXT_RSA_DES_192_CBC3_SHA	/*"DES-CBC3-SHA"*/,	168, 168 },
+    { SSL3_TXT_DH_DSS_DES_40_CBC_SHA	/*"EXP-DH-DSS-DES-CBC-SHA"*/, 56, 40 },
+    { SSL3_TXT_DH_DSS_DES_64_CBC_SHA	/*"DH-DSS-DES-CBC-SHA"*/, 56, 56 },
+    { SSL3_TXT_DH_DSS_DES_192_CBC3_SHA 	/*"DH-DSS-DES-CBC3-SHA"*/, 168, 168 },
+    { SSL3_TXT_DH_RSA_DES_40_CBC_SHA	/*"EXP-DH-RSA-DES-CBC-SHA"*/, 56, 40 },
+    { SSL3_TXT_DH_RSA_DES_64_CBC_SHA	/*"DH-RSA-DES-CBC-SHA"*/, 56, 56 },
+    { SSL3_TXT_DH_RSA_DES_192_CBC3_SHA 	/*"DH-RSA-DES-CBC3-SHA"*/, 168, 168 },
+    { SSL3_TXT_EDH_DSS_DES_40_CBC_SHA	/*"EXP-EDH-DSS-DES-CBC-SHA"*/, 56, 40 },
+    { SSL3_TXT_EDH_DSS_DES_64_CBC_SHA	/*"EDH-DSS-DES-CBC-SHA"*/, 56, 56 },
+    { SSL3_TXT_EDH_DSS_DES_192_CBC3_SHA	/*"EDH-DSS-DES-CBC3-SHA"*/, 168, 168 },
+    { SSL3_TXT_EDH_RSA_DES_40_CBC_SHA	/*"EXP-EDH-RSA-DES-CBC"*/, 56, 40 },
+    { SSL3_TXT_EDH_RSA_DES_64_CBC_SHA	/*"EDH-RSA-DES-CBC-SHA"*/, 56, 56 },
+    { SSL3_TXT_EDH_RSA_DES_192_CBC3_SHA	/*"EDH-RSA-DES-CBC3-SHA"*/, 168, 168 },
+    { SSL3_TXT_ADH_RC4_40_MD5		/*"EXP-ADH-RC4-MD5"*/,	128, 40 },
+    { SSL3_TXT_ADH_RC4_128_MD5		/*"ADH-RC4-MD5"*/,	128, 128 },
+    { SSL3_TXT_ADH_DES_40_CBC_SHA	/*"EXP-ADH-DES-CBC-SHA"*/, 128, 40 },
+    { SSL3_TXT_ADH_DES_64_CBC_SHA	/*"ADH-DES-CBC-SHA"*/,	56, 56 },
+    { SSL3_TXT_ADH_DES_192_CBC_SHA	/*"ADH-DES-CBC3-SHA"*/, 168, 168 },
+    { SSL3_TXT_FZA_DMS_NULL_SHA		/*"FZA-NULL-SHA"*/,	0, 0 },
+    { SSL3_TXT_FZA_DMS_FZA_SHA		/*"FZA-FZA-CBC-SHA"*/,	-1, -1 },
+    { SSL3_TXT_FZA_DMS_RC4_SHA		/*"FZA-RC4-SHA"*/,	 128, 128 },
+    { SSL2_TXT_DES_64_CFB64_WITH_MD5_1	/*"DES-CFB-M1"*/,	56, 56 },
+    { SSL2_TXT_RC2_128_CBC_WITH_MD5	/*"RC2-CBC-MD5"*/,	128, 128 },
+    { SSL2_TXT_DES_64_CBC_WITH_MD5	/*"DES-CBC-MD5"*/,	56, 56 },
+    { SSL2_TXT_DES_192_EDE3_CBC_WITH_MD5/*"DES-CBC3-MD5"*/,	168, 168 },
+    { SSL2_TXT_RC4_64_WITH_MD5		/*"RC4-64-MD5"*/,	64, 64 },
+    { SSL2_TXT_NULL			/*"NULL"*/,		0, 0 },
+    { NULL, 0, 0 }
+    };
+
+int SSLFixups(request_rec *r)
     {
     table *e=r->subprocess_env;
     int keysize=0;
     int secretkeysize=0;
     char buf[MAX_STRING_LEN];
     char *cipher;
+    int n;
+    X509 *xs;
 
     SSLConfigRec *pConfig=(SSLConfigRec *)
       get_module_config(r->server->module_config,&ssl_module);
 
     /* Check to see if SSL is on */
-    if(pConfig->bDisabled)
+    if(pConfig->bDisabled || OPTIONAL_SSL)
 	return DECLINED;
 
     cipher=SSL_get_cipher(r->connection->client->ssl);
     table_set(e,"HTTPS","on");
     table_set(e,"HTTPS_CIPHER",cipher);
+    table_set(e,"SSL_CIPHER",cipher);
+    table_set(e,"SSL_PROTOCOL_VERSION",
+	      SSL_get_version(r->connection->client->ssl));
+    table_set(e,"SSL_SSLEAY_VERSION", SSLeay_version(SSLEAY_VERSION));
 
-    if(!strcmp(cipher,SSL_TXT_NULL_WITH_MD5))
-	keysize=secretkeysize=0;
-    else if(!strcmp(cipher,SSL_TXT_RC4_128_WITH_MD5))
-	keysize=secretkeysize=128;
-    else if(!strcmp(cipher,SSL_TXT_RC4_128_EXPORT40_WITH_MD5))
-	{ keysize=128; secretkeysize=40; }
-    else if(!strcmp(cipher,SSL_TXT_RC2_128_CBC_WITH_MD5))
-	keysize=secretkeysize=128;
-    else if(!strcmp(cipher,SSL_TXT_RC2_128_CBC_EXPORT40_WITH_MD5))
-	{ keysize=128; secretkeysize=40; }
-    else if(!strcmp(cipher,SSL_TXT_IDEA_128_CBC_WITH_MD5))
-	keysize=secretkeysize=128;
-    else if(!strcmp(cipher,SSL_TXT_DES_64_CBC_WITH_MD5))
-	keysize=secretkeysize=64;
-    else if(!strcmp(cipher,SSL_TXT_DES_64_CBC_WITH_SHA))
-	keysize=secretkeysize=64;
-    else if(!strcmp(cipher,SSL_TXT_DES_192_EDE3_CBC_WITH_MD5))
-	keysize=secretkeysize=192;
-    else if(!strcmp(cipher,SSL_TXT_DES_192_EDE3_CBC_WITH_SHA))
-	keysize=secretkeysize=192;
-    else if(!strcmp(cipher,SSL_TXT_DES_64_CFB64_WITH_MD5_1))
-	keysize=secretkeysize=64;
-    else if(!strcmp(cipher,SSL_TXT_NULL))
-	keysize=secretkeysize=0;
+    for(n=0 ; aCipherspecs[n].szName ; ++n)
+	if(!strcmp(cipher,aCipherspecs[n].szName))
+	    {
+	    keysize=aCipherspecs[n].nKeySize;
+	    secretkeysize=aCipherspecs[n].nSecretKeySize;
+	    break;
+	    }
 
     sprintf(buf,"%d",keysize);
     table_set(e,"HTTPS_KEYSIZE",buf);
@@ -247,9 +344,20 @@ int SSLFixups (request_rec *r)
     table_set(e,"HTTPS_SECRETKEYSIZE",buf);
 
     if(r->connection->client->szClientX509)
-	table_set(e,"SSL_CLIENT_DN",r->connection->client->szClientX509);
+	{
+	ExpandCert(e,"SSL_CLIENT_",r->connection->client->szClientX509);
+	xs=SSL_get_peer_certificate(r->connection->client->ssl);
+	ExpandCert(e,"SSL_CLIENT_I",X509_NAME_oneline(X509_get_issuer_name(xs),
+						      NULL,0));
+	}
 
-    return OK;
+    xs=SSL_get_certificate(r->connection->client->ssl);
+    ExpandCert(e,"SSL_SERVER_",X509_NAME_oneline(X509_get_subject_name(xs),
+						 NULL,0));
+    ExpandCert(e,"SSL_SERVER_I",X509_NAME_oneline(X509_get_issuer_name(xs),
+						  NULL,0));
+
+    return DECLINED;
     }
 
 static void SSLLogError(server_rec *s)
@@ -266,6 +374,22 @@ static void SSLLogError(server_rec *s)
       }
     }
 
+/* for debugging */
+static void SSLLogErrorToFile(FILE *f)
+    {
+    unsigned long l;
+    char buf[MAX_STRING_LEN];
+
+    /* Print out error messages */
+
+    while((l=ERR_get_error()))
+      {
+	ERR_error_string(l,buf);
+	fputs(buf,f);
+	fputc('\n',f);
+      }
+    }
+
 static void *CreateSSLServerConfig(pool *p,server_rec *s)
     {
     SSLConfigRec *rec=pcalloc(p,sizeof(SSLConfigRec));
@@ -277,6 +401,7 @@ static void *CreateSSLServerConfig(pool *p,server_rec *s)
     rec->px509Certificate=NULL;
     rec->prsaKey=NULL;
     rec->bFakeBasicAuth=FALSE;
+    rec->tSessionCacheTimeout=0;
     return rec;
     }
 
@@ -307,12 +432,26 @@ static void InitSSL()
     {
     char *CAfile=NULL,*CApath=NULL;
 
+#if CACHE_SESSIONS
+    if(!szCacheServerPort)
+	{
+	fprintf(stderr,"Required SSLCacheServerPort missing\n");
+	exit(1);
+	}
+    if(!szCacheServerPath)
+	{
+	fprintf(stderr,"Required SSLCacheServerPath missing\n");
+	exit(1);
+	}
+#endif
+
     SSL_load_error_strings();
     ERR_load_crypto_strings();
 #if SSLEAY_VERSION_NUMBER >= 0x0800
     SSLeay_add_ssl_algorithms();
 #else
-    SSL_debug("/tmp/ssldebug");
+    /* only enable this with care: it can be used for /tmp symlink attacks.
+    SSL_debug("/tmp/ssldebug");*/
 #endif
     }
 
@@ -324,7 +463,11 @@ static RSA *TmpRSACallback(SSL *pSSL,int nExport)
     static RSA *pRSA=NULL;
 
     if (pRSA == NULL)
+#if SSLEAY_VERSION_NUMBER >= 0x0900
+	pRSA=RSA_generate_key(512,RSA_F4,NULL,NULL);
+#else
 	pRSA=RSA_generate_key(512,RSA_F4,NULL);
+#endif
     return pRSA;
     }
 #endif
@@ -339,42 +482,48 @@ int ApacheSSLVerifyCallback(int ok,X509_STORE_CTX *ctx)
 int ApacheSSLVerifyCallback(int ok,X509 *xs,X509 *xi,int depth,int error)
     {
 #endif
-    char *s;
-    SSLConfigRec *pConfig=get_module_config(SSLVerifyConn->server->module_config,&ssl_module);
+    SSLConfigRec *pConfig=get_module_config(g_pCurrentConnection->server->module_config,&ssl_module);
+
+#if DEBUG_VERIFY
+      {
+      char *s;
 
 #if SSLEAY_VERSION_NUMBER < 0x0800
-    s=(char *)X509_NAME_oneline(X509_get_subject_name(xs));
+      s=X509_NAME_oneline(X509_get_subject_name(xs));
 #else
-    s=(char *)X509_NAME_oneline(X509_get_subject_name(xs),NULL,0);
+      s=X509_NAME_oneline(X509_get_subject_name(xs),NULL,0);
 #endif
 
-    if(s == NULL)
-	{
-	ERR_print_errors_fp(pConfig->fileLogFile);
-	return(0);
-	}
-    if(depth == 0)
-	SSLVerifyConn->client->szClientX509=pstrdup(SSLVerifyConn->pool,s);
+      if(s == NULL)
+	  {
+	  ERR_print_errors_fp(pConfig->fileLogFile);
+	  return(0);
+	  }
 
-    fprintf(pConfig->fileLogFile,"depth=%d %s\n",depth,s);
-    free(s);
-    if(error == X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY)
+      fprintf(pConfig->fileLogFile,"depth=%d %s\n",depth,s);
+      free(s);
+      }
+#endif
+
+    if(error == X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY || error == X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE)
 	{
+	char *s;
+
 	if(pConfig->nVerifyClient == VERIFY_OPTIONAL_NO_CA)
 	    {
 	    fprintf(pConfig->fileLogFile,"no issuer, returning OK\n");
 	    return TRUE;
 	    }
 #if SSLEAY_VERSION_NUMBER < 0x0800
-	s=(char *)X509_NAME_oneline(X509_get_issuer_name(xs));
+	s=X509_NAME_oneline(X509_get_issuer_name(xs));
 #else
-	s=(char *)X509_NAME_oneline(X509_get_issuer_name(xs),NULL,0);
+	s=X509_NAME_oneline(X509_get_issuer_name(xs),NULL,0);
 #endif
 	if(s == NULL)
 	    {
 	    fprintf(pConfig->fileLogFile,"verify error\n");
 	    ERR_print_errors_fp(pConfig->fileLogFile);
-	    SSLLogError(SSLVerifyConn->server);
+	    SSLLogError(g_pCurrentConnection->server);
 	    return(0);
 	    }
 	fprintf(pConfig->fileLogFile,"issuer= %s\n",s);
@@ -384,13 +533,13 @@ int ApacheSSLVerifyCallback(int ok,X509 *xs,X509 *xi,int depth,int error)
 	{
 	fprintf(pConfig->fileLogFile,"verify error:num=%d:%s\n",error,
 		X509_verify_cert_error_string(error));
-	SSLLogError(SSLVerifyConn->server);
-	SSLVerifyConn->client->szClientX509=NULL;
+	SSLLogError(g_pCurrentConnection->server);
+	g_pCurrentConnection->client->szClientX509=NULL;
 	}
     if(depth >= pConfig->nVerifyDepth)
 	{
 	fprintf(pConfig->fileLogFile,"Verify depth exceeded\n");
-	log_error("Verify depth exceeded",SSLVerifyConn->server);
+	log_error("Verify depth exceeded",g_pCurrentConnection->server);
 	ok=0;
 	}
     fprintf(pConfig->fileLogFile,"verify return:%d\n",ok);
@@ -415,6 +564,110 @@ static int VerifyFlags(SSLConfigRec *pConfig)
     return nVerify;
     }
 
+#if CACHE_SESSIONS
+
+#define MAX_SESSION_DER	100000
+
+static void SendSessionToServer(SSL_SESSION *pSession)
+    {
+    uchar buf[MAX_SESSION_DER];
+    Cache *p;
+    uchar *t;
+    int nLength;
+    SSLConfigRec *pConfig;
+    time_t tExpiresAt;
+
+    if(LocalCacheFind(pSession->session_id,pSession->session_id_length))
+	return;
+
+    pConfig=get_module_config(g_pCurrentConnection->server->module_config,&ssl_module);
+    tExpiresAt=time(NULL)+pConfig->tSessionCacheTimeout;
+
+    t=malloc(pSession->session_id_length);
+    memcpy(t,pSession->session_id,pSession->session_id_length);
+    p=LocalCacheAdd(t,pSession->session_id_length,tExpiresAt);
+
+    t=buf;
+    nLength=i2d_SSL_SESSION(pSession,&t);
+    assert(nLength > 0);
+    assert(nLength <= MAX_SESSION_DER);
+
+    p->aucData=malloc(nLength);
+    memcpy(p->aucData,buf,nLength);
+    p->nData=nLength;
+
+    GlobalCacheAdd(p->aucKey,p->nKey,p->aucData,p->nData,tExpiresAt);
+    }
+
+static SSL_SESSION *GetSessionFromServer(uchar *aucSessionID,int nLength)
+    {
+    int n;
+    SSL_SESSION *pSession;
+    Cache *p;
+    uchar *t,*t2;
+    int nData;
+
+    p=LocalCacheFind(aucSessionID,nLength);
+    if(!p)
+	{
+	time_t tExpiresAt;
+
+	t=GlobalCacheGet(aucSessionID,nLength,&nData,&tExpiresAt);
+	if(!t)
+	    return NULL;
+	t2=malloc(nLength);
+	memcpy(t2,aucSessionID,nLength);
+	p=LocalCacheAdd(t2,nLength,tExpiresAt);
+	p->aucData=t;
+	p->nData=nData;
+	}
+    else
+	{
+	t=p->aucData;
+	nData=p->nData;
+	}
+    pSession=d2i_SSL_SESSION(NULL,&t,nData);
+    if(!pSession)
+	SSLLogErrorToFile(stderr);
+    assert(pSession);
+    return pSession;
+    }
+
+static int NewSessionCacheCallback(SSL *pSSL,SSL_SESSION *pNew)
+    {
+#if DEBUG_SESSIONS
+    int n;
+
+    fprintf(stderr,"NewSessionCacheCallback(");
+    for(n=0 ; n < pNew->session_id_length ; ++n)
+	fprintf(stderr,"%02x",pNew->session_id[n]);
+    fprintf(stderr,")\n");
+#endif
+
+    SendSessionToServer(pNew);
+    return 0;
+    }
+
+static SSL_SESSION *GetSessionCacheCallback(SSL *pSSL,unsigned char *aucSessionID,int nLength,int *pCopy)
+    {
+    SSL_SESSION *pSession,*p2;
+
+#if DEBUG_SESSIONS
+    int n;
+
+    fprintf(stderr,"GetSessionCacheCallback(");
+    for(n=0 ; n < nLength ; ++n)
+	fprintf(stderr,"%02x",aucSessionID[n]);
+    fprintf(stderr,")\n");
+#endif
+    *pCopy=0;
+    pSession=GetSessionFromServer(aucSessionID,nLength);
+
+    return pSession;
+    }
+
+#endif /* CACHE_SESSIONS */
+
 static void GetCertificateAndKey(server_rec *s,SSLConfigRec *pConfig)
     {
     char buf[1024],prompt[1024];
@@ -440,6 +693,11 @@ static void GetCertificateAndKey(server_rec *s,SSLConfigRec *pConfig)
     SSL_CTX_set_verify(pConfig->pSSLCtx,VerifyFlags(pConfig),ApacheSSLVerifyCallback);
 #endif
 
+#if CACHE_SESSIONS
+    SSL_CTX_sess_set_new_cb(pConfig->pSSLCtx,NewSessionCacheCallback);
+    SSL_CTX_sess_set_get_cb(pConfig->pSSLCtx,GetSessionCacheCallback);
+#endif
+
     if(pConfig->szReqCiphers != NULL)
 	{
 	if(!SSL_CTX_set_cipher_list(pConfig->pSSLCtx,pConfig->szReqCiphers))
@@ -450,15 +708,17 @@ static void GetCertificateAndKey(server_rec *s,SSLConfigRec *pConfig)
 	    }
 	}
 
-    if(!SSL_CTX_load_verify_locations(pConfig->pSSLCtx,
-				  pConfig->szCACertificateFile,
-				  pConfig->szCACertificatePath)
+    if(((pConfig->szCACertificateFile || pConfig->szCACertificatePath)
+	&& !SSL_CTX_load_verify_locations(pConfig->pSSLCtx,
+					   pConfig->szCACertificateFile,
+					   pConfig->szCACertificatePath)
+	)
        || !SSL_CTX_set_default_verify_paths(pConfig->pSSLCtx))
-       {
-       fprintf(stderr,"error seting default verify locations\n");
-       ERR_print_errors_fp(stderr);
-       exit(1);
-       }
+	{
+	fprintf(stderr,"error setting verify locations\n");
+	ERR_print_errors_fp(stderr);
+	exit(1);
+	}
 
     f=fopen(pConfig->szCertificateFile,"r");
     if(!f)
@@ -520,8 +780,41 @@ static void GetCertificateAndKey(server_rec *s,SSLConfigRec *pConfig)
 #endif
     }
 
+#if CACHE_SESSIONS
+static void CacheLauncher(void *cmd)
+    {
+    char a1[1024];
+    extern uid_t user_id;
+    extern server_rec *server_conf;	/* Since the cache is global, use the main server's log files */
+
+    error_log2stderr(server_conf);
+    dup2(STDERR_FILENO,STDOUT_FILENO);
+    sprintf(a1,"%d",user_id);
+    execl(cmd,"gcache",a1,szCacheServerPort,NULL);
+    assert(0);
+    }
+
+static void LaunchCache(pool *pPool)
+    {
+    int pid;
+
+    printf("Launching... %s\n",szCacheServerPath);
+    pid=spawn_child_err(pPool,CacheLauncher,szCacheServerPath,kill_after_timeout,NULL,NULL,NULL);
+    printf("pid=%d\n",pid);
+    }
+#endif
+
+void LaunchGlobalCache(pool *pPool)
+    {
+#if CACHE_SESSIONS
+    LaunchCache(pPool);
+#endif
+    }
+
 static void InitSSLServer(server_rec *s,pool *p)
     {
+    char *szLogFile;
+
     InitSSL();
     for ( ; s ; s=s->next)
 	{
@@ -535,7 +828,13 @@ static void InitSSLServer(server_rec *s,pool *p)
 		    s->server_hostname,s->port);
 	    continue;
 	    }
-
+#if CACHE_SESSIONS
+	if(!pConfig->tSessionCacheTimeout)
+	    {
+	    fprintf(stderr,"SSLSessionCacheTimeout not set\n");
+	    exit(1);
+	    }
+#endif
 	if(!s->port)
 	    s->port=HTTPS_PORT;
 
@@ -555,8 +854,9 @@ static void InitSSLServer(server_rec *s,pool *p)
 	    fprintf(stderr,"Required SSLLogFile missing\n");
 	    exit(1);
 	    }
-	
-	pConfig->fileLogFile=fopen(pConfig->szLogFile,"a");
+
+	szLogFile=server_root_relative(p,pConfig->szLogFile);
+	pConfig->fileLogFile=fopen(szLogFile,"a");
 	if(!pConfig->fileLogFile)
 	    {
 	    perror(pConfig->szLogFile);
@@ -568,6 +868,24 @@ static void InitSSLServer(server_rec *s,pool *p)
 	}
     }
 
+#if CACHE_SESSIONS
+static const char *SSLCacheServerPort(cmd_parms *cmd,char *struct_ptr,char *arg)
+    {
+    szCacheServerPort=server_root_relative(cmd->pool,arg);
+    InitGlobalCache(szCacheServerPort);
+
+    return NULL;
+    }
+
+static const char *SSLCacheServerPath(cmd_parms *cmd,char *struct_ptr,char *arg)
+    {
+    szCacheServerPath=server_root_relative(cmd->pool,arg);
+
+    return NULL;
+    }
+#endif
+    
+
 static const char *set_server_string_slot (cmd_parms *cmd,char *struct_ptr,char *arg)
     {
     /* This one should be pretty generic... */
@@ -598,9 +916,22 @@ static const char *set_server_bool_slot (cmd_parms *cmd,char *struct_ptr)
     return NULL;
     }
 
+static const char *set_server_bool_slot_false (cmd_parms *cmd,char *struct_ptr)
+    {
+  /* This one should be pretty generic... */
+    char *pConfig=get_module_config(cmd->server->module_config,&ssl_module);
+  
+    int offset=(int)cmd->info; 
+    *(BOOL *)(pConfig+offset)=FALSE;
+    return NULL;
+    }
+
 command_rec ssl_cmds[]=
     {
     { "SSLDisable",set_server_bool_slot,
+      (void *)XtOffsetOf(SSLConfigRec,bDisabled),RSRC_CONF,NO_ARGS,
+      "Disable SSL" },
+    { "SSLEnable",set_server_bool_slot_false,
       (void *)XtOffsetOf(SSLConfigRec,bDisabled),RSRC_CONF,NO_ARGS,
       "Enable SSL" },
     { "SSLCertificateFile",set_server_string_slot,
@@ -632,9 +963,18 @@ command_rec ssl_cmds[]=
       "Colon-delimited list of required ciphers" },
     /* Per Directory */
     { "SSLRequireCipher",SSLRequireCipher,NULL,OR_FILEINFO,ITERATE,
-      "add a cipher to the per directory list of required ciphers" },
+      "Add a cipher to the per directory list of required ciphers" },
     { "SSLBanCipher",SSLBanCipher,NULL,OR_FILEINFO,ITERATE,
-      "add a cipher to the per directory list of banned ciphers" },
+      "Add a cipher to the per directory list of banned ciphers" },
+#if CACHE_SESSIONS
+    { "SSLCacheServerPort",SSLCacheServerPort,NULL,RSRC_CONF,TAKE1,
+      "Cache server TCP port number" },
+    { "SSLSessionCacheTimeout",set_server_int_slot,
+      (void *)XtOffsetOf(SSLConfigRec,tSessionCacheTimeout),RSRC_CONF,TAKE1,
+      "Lifetime in seconds of session cache objects" },
+    { "SSLCacheServerPath",SSLCacheServerPath,NULL,RSRC_CONF,TAKE1,
+      "Path of the global cache server" },
+#endif
     { NULL },
     };
 
@@ -677,6 +1017,8 @@ static int FakeBasicAuth(request_rec *r)
     SSLConfigRec *pConfig=get_module_config(r->server->module_config,&ssl_module);
     char b1[MAX_STRING_LEN],b2[MAX_STRING_LEN];
   
+    if(OPTIONAL_SSL)
+	return DECLINED;
     if(!pConfig->bFakeBasicAuth)
 	return DECLINED;
     if(r->connection->user)
@@ -736,12 +1078,12 @@ void ApacheSSLSetupVerify(conn_rec *conn)
 
     conn->client->szClientX509=NULL;
 
-    SSLVerifyConn=conn;
+    g_pCurrentConnection=conn;
 
     conn->client->nVerifyError=X509_V_OK;
 
-/* Why call this twice?? Ben */  
 #if SSLEAY_VERSION_NUMBER < 0x0800
+/* Why call this twice?? Ben */  
     SSL_set_verify(conn->client->ssl,nVerify,ApacheSSLVerifyCallback);
 #endif
     if(!ApacheSSLSetCertStuff(conn))
@@ -750,16 +1092,33 @@ void ApacheSSLSetupVerify(conn_rec *conn)
 	log_error("ApacheSSLSetCertStuff failed",conn->server);
 	exit(1);
 	}
-#if SSLEAY_VERSION_NUMBER >= 0x0800
-#else
+#if SSLEAY_VERSION_NUMBER < 0x0800
     SSL_set_verify(conn->client->ssl,nVerify,ApacheSSLVerifyCallback);
 #endif
     }
 
+#if FREE_SESSION
+/*
+To make this work, SSL_free needs to be patched. Remove from SSL_free:
+
+	if (s->ctx) SSL_CTX_free(s->ctx);
+
+but there's more, which I haven't found.
+*/
+
+static void CloseSSL(void *_ssl)
+    {
+    SSL *ssl=_ssl;
+printf("Closing\n");
+    SSL_free(_ssl);
+printf("Closed\n");
+    }
+#endif
+
 int ApacheSSLSetupConnection(conn_rec * conn)
     {
-    server_rec * srvr=conn->server;
-    BUFF * fb=conn->client;
+    server_rec *srvr=conn->server;
+    BUFF *fb=conn->client;
     SSLConfigRec *pConfig=get_module_config(srvr->module_config,
 					      &ssl_module);
     char *cert_file=pConfig->szCertificateFile;
@@ -771,31 +1130,14 @@ int ApacheSSLSetupConnection(conn_rec * conn)
 	return TRUE;
 	}
     
-    SSLVerifyConn=conn;
+    g_pCurrentConnection=conn;
 
     fb->ssl=SSL_new(pConfig->pSSLCtx);
     SSL_set_fd(fb->ssl,fb->fd);
-
-#if 0
-    if (cert_file != NULL) {
-	if (SSL_use_certificate(fb->ssl,pConfig->px509Certificate) <= 0) {
-	    fprintf(pConfig->fileLogFile,"unable to set certificate\n");
-	    ERR_print_errors_fp(pConfig->fileLogFile);
-	    SSLLogError(srvr);
-	    fb->flags |= B_EOF | B_EOUT;
-	    return FALSE;
-	}
-
-	if(SSL_use_RSAPrivateKey(fb->ssl,pConfig->prsaKey) <= 0) {
-	    fprintf(pConfig->fileLogFile,"unable to set private key\n");
-	    ERR_print_errors_fp(pConfig->fileLogFile);
-	    SSLLogError(srvr);
-	    fb->flags |= B_EOF | B_EOUT;
-	    return FALSE;
-	}
-
-    }
+#if FREE_SESSION
+    register_cleanup(conn->pool,fb->ssl,CloseSSL,CloseSSL);
 #endif
+
     ApacheSSLSetupVerify(conn);
 
     while(!SSL_is_init_finished(fb->ssl))
@@ -817,6 +1159,19 @@ int ApacheSSLSetupConnection(conn_rec * conn)
 	    return FALSE;
 	    }
 
+	if(pConfig->nVerifyClient != VERIFY_NONE)
+	    {
+	    char *s;
+	    X509 *xs;
+	    xs=SSL_get_peer_certificate(fb->ssl);
+	    if(xs)
+		{
+		s=X509_NAME_oneline(X509_get_subject_name(xs),NULL,0);
+		conn->client->szClientX509=pstrdup(conn->pool,s);
+		free(s);
+		}
+	    }
+
 	if(pConfig->nVerifyClient == VERIFY_REQUIRE && !conn->client->szClientX509)
 	    {
 	    log_error("No client certificate",conn->server);
@@ -830,7 +1185,7 @@ int ApacheSSLSetupConnection(conn_rec * conn)
     /* This should be safe.... so I'll use it */
     SSL_set_read_ahead(fb->ssl,1);
 
-    SSLVerifyConn=NULL;
+    g_pCurrentConnection=NULL;
 
     return TRUE;
     }
@@ -852,4 +1207,5 @@ module ssl_module =
     NULL,			/* type_checker */
     SSLFixups,			/* fixups */
     NULL,			/* logger */
-    };
+    }
+;
